/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.semantik;

import static de.ozgcloud.eingang.common.formdata.FormDataTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.router.VorgangService;
import de.ozgcloud.eingang.semantik.enginebased.EngineBasedSemantikAdapter;
import de.ozgcloud.eingang.semantik.formbased.FormBasedSemantikAdapter;

class SemantikAdapterTest {

	@Spy
	@InjectMocks
	private SemantikAdapter adapter;

	@Mock
	private List<EngineBasedSemantikAdapter> engineAdapters;

	@Mock
	private EngineBasedSemantikAdapter engineAdapter;
	@Mock
	private FormBasedSemantikAdapter formAdapter;
	@Mock
	private VorgangService vorgangService;

	@DisplayName("get responsible engine adapters")
	@Nested
	class TestGetResponsibleEngineAdapters {
		@Mock
		private EngineBasedSemantikAdapter unresponsibleEngineAdapter;
		@Mock
		private FormData formData;

		@BeforeEach
		void mock() {
			when(unresponsibleEngineAdapter.isResponsible(formData)).thenReturn(false);
			when(engineAdapter.isResponsible(formData)).thenReturn(true);
			when(engineAdapters.stream()).thenReturn(Stream.of(engineAdapter, unresponsibleEngineAdapter));
		}

		@DisplayName("should return responsible")
		@Test
		void shouldReturnResponsible() {
			var foundAdapters = adapter.getResponsibleEngineAdapters(formData);

			assertThat(foundAdapters).isEqualTo(List.of(engineAdapter));
		}
	}

	@Nested
	class TestProcessFormData {

		@Mock
		private FormData formData;
		@Mock
		private FormData engineAdapterResponse;
		@Mock
		private FormData formAdapterResponse;

		@BeforeEach
		void mockEngineAdapter() {
			doReturn(List.of(engineAdapter)).when(adapter).getResponsibleEngineAdapters(formData);
			when(engineAdapter.parseFormData(any())).thenReturn(engineAdapterResponse);
			when(formAdapter.parseFormData(any())).thenReturn(formAdapterResponse);
			when(vorgangService.createVorgang(any())).thenReturn(VORGANG_ID);
		}

		@Test
		void shouldCallFindEngineAdapter() {
			adapter.processFormData(formData);

			verify(adapter).getResponsibleEngineAdapters(formData);
		}

		@Test
		void shouldCallEngineAdapter() {
			adapter.processFormData(formData);

			verify(engineAdapter).parseFormData(formData);
		}

		@Test
		void shouldCallFormAdapter() {
			adapter.processFormData(formData);

			verify(formAdapter).parseFormData(engineAdapterResponse);
		}

		@Test
		void shouldCallVorgangRemoteService() {
			adapter.processFormData(formData);

			verify(vorgangService).createVorgang(formAdapterResponse);
		}

		@DisplayName("should return first vorgang id")
		@Test
		void shouldReturnFirstVorgangId() {
			var vorgangId = adapter.processFormData(formData);

			assertThat(vorgangId).isEqualTo(VORGANG_ID);
		}
	}
}
