package de.ozgcloud.eingang.xta;

import static de.ozgcloud.eingang.xta.XtaService.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;
import de.ozgcloud.eingang.common.vorgang.VorgangNummerSupplier;
import de.ozgcloud.eingang.xdomea.XdomeaMessageDataMapper;

class XtaServiceTest {

	private static final String MESSAGE_TYPE_OTHER = "MESSAGE_TYPE_OTHER";

	@Spy
	@InjectMocks
	private XtaService service;

	@Mock
	private XtaRemoteService remoteService;
	@Mock
	private XtaMessageMapper mapper;
	@Mock
	private VorgangNummerSupplier vorgangNummerSupplier;

	@Mock
	private XtaIncomingFilesMapper incomingFilesMapper;

	@Mock
	private XdomeaMessageDataMapper xdomeaMessageDataMapper;

	@Mock
	private XtaCurrentIdentifierService currentIdentifierService;

	@Mock
	private FormData formData;

	@Mock
	private FormData formData2;

	@Nested
	class TestGetMessagesAsFormData {

		private final XtaMessageMetaData messageMetaData = XtaMessageMetaDataTestFactory.create();
		private final XtaMessage message = XtaMessageTestFactory.create();

		@DisplayName("with one message")
		@Nested
		class TestWithOneMessage {
			@BeforeEach
			void setup() {
				doReturn(Stream.of(messageMetaData)).when(service).createXtaMessageStream();
			}

			@Test
			void shouldCallCreateStream() {
				service.getMessages();

				verify(service).createXtaMessageStream();
			}

			@Test
			void shouldCallFilterByMessageType() {
				setupMocks();

				service.getMessages().toList();

				verify(service).isSupportedMessageType(messageMetaData);
			}

			@Test
			void shouldCallGetFormDataIfNoRuntimeException() {
				setupMocks();
				doReturn(true).when(service).isSupportedMessageType(messageMetaData);

				service.getMessages().toList();

				verify(service).getFormData(messageMetaData);
			}

			@Test
			void shouldNotCallGetFormDataIfNoRuntimeException() {
				doReturn(false).when(service).isSupportedMessageType(messageMetaData);

				service.getMessages().toList();

				verify(service, never()).getFormDataIfNoRuntimeException(any());
			}
		}

		@DisplayName("with multiple messages")
		@Nested
		class TestWithMultipleMessages {
			private final XtaMessageMetaData messageMetaData2 = XtaMessageMetaDataTestFactory.createBuilder()
					.messageId(XtaMessageId.from("messageId2"))
					.build();
			private final XtaMessageMetaData messageMetaData3 = XtaMessageMetaDataTestFactory.createBuilder()
					.messageId(XtaMessageId.from("messageId3"))
					.build();

			@BeforeEach
			void setup() {
				doReturn(Stream.of(messageMetaData, messageMetaData2, messageMetaData3)).when(service).createXtaMessageStream();
			}

			@DisplayName("should return stream of messages")
			@Test
			void shouldReturnStreamOfMessages() {
				doReturn(true).when(service).isSupportedMessageType(messageMetaData);
				doReturn(false).when(service).isSupportedMessageType(messageMetaData3);
				doReturn(true).when(service).isSupportedMessageType(messageMetaData2);
				doReturn(Optional.of(formData)).when(service).getFormDataIfNoRuntimeException(messageMetaData);
				doReturn(Optional.of(formData2)).when(service).getFormDataIfNoRuntimeException(messageMetaData2);

				var result = service.getMessages().toList();

				assertThat(result).containsExactly(formData, formData2);
			}
		}


		private void setupMocks() {
			var testFormData = FormDataTestFactory.create();
			when(mapper.toFormData(any(), any(), eq(vorgangNummerSupplier))).thenReturn(testFormData);
			when(remoteService.getMessage(any(XtaMessageId.class))).thenReturn(message);
		}
	}

	@Nested
	class TestIsSupportedMessageType {

		@Test
		void shouldAcceptGeschaeftsgang() {
			var geschaeftsgangMetaData = XtaMessageMetaDataTestFactory.create();

			assertThat(service.isSupportedMessageType(geschaeftsgangMetaData)).isTrue();
		}

		@Test
		void shouldAcceptFIM() {
			var fimMetaData = XtaMessageMetaDataTestFactory.createBuilder()
					.messageType(FIM_MESSAGE_TYPE_PREFIX + "34355")
					.build();

			assertThat(service.isSupportedMessageType(fimMetaData)).isTrue();
		}

		@Test
		void shouldNotAcceptOtherMessageType() {
			var metaDataDFoerder = XtaMessageMetaDataTestFactory.createBuilder().messageType(MESSAGE_TYPE_OTHER).build();

			assertThat(service.isSupportedMessageType(metaDataDFoerder)).isFalse();
		}
	}

	@DisplayName("get form data if no runtime exception")
	@Nested
	class TestGetFormDataIfNoRuntimeException {

		@Mock
		XtaMessageMetaData messageMetaData;

		@Mock
		FormData formData;

		@DisplayName("should call get formdata")
		@Test
		void shouldCallGetFormdata() {
			service.getFormDataIfNoRuntimeException(messageMetaData);

			verify(service).getFormData(messageMetaData);
		}

		@DisplayName("with exception")
		@Nested
		class TestWithException {
			@DisplayName("should return empty")
			@Test
			void shouldReturnEmpty() {
				doThrow(new RuntimeException("test-error")).when(service).getFormData(any());

				var result = service.getFormDataIfNoRuntimeException(messageMetaData);

				assertThat(result).isEmpty();
			}
		}

		@DisplayName("without exception")
		@Nested
		class TestWithoutException {
			@DisplayName("should return")
			@Test
			void shouldReturn() {
				doReturn(formData).when(service).getFormData(any());

				var result = service.getFormDataIfNoRuntimeException(messageMetaData);

				assertThat(result).contains(formData);
			}
		}
	}

	@Nested
	class TestGetFormData {

		@Mock
		private FormData formData;

		private XtaMessage message = XtaMessageTestFactory.create();
		private XtaMessageMetaData messageMetaData = XtaMessageMetaDataTestFactory.create();
		private RepresentationsAttachmentsPair classification;
		private FormData mappedFormData = FormDataTestFactory.create();

		@Captor
		private ArgumentCaptor<XtaMessageMetaData> messageMetaDataCaptor;

		@Captor
		private ArgumentCaptor<RepresentationsAttachmentsPair> classificationCaptor;

		@BeforeEach
		void init() {
			classification = RepresentationsAttachmentsPairTestFactory.create();
			var incomingFiles = List.of(IncomingFileTestFactory.create(), IncomingFileTestFactory.create());
			when(incomingFilesMapper.toIncomingFiles(message.getMessageFiles())).thenReturn(incomingFiles);
			doReturn(classification).when(service).getRepresentationsAttachmentsPair(messageMetaData, incomingFiles);
		}

		@BeforeEach
		void mockMessageMapping() {
			when(remoteService.getMessage(any(XtaMessageId.class))).thenReturn(message);
			when(mapper.toFormData(any(), any(), any())).thenReturn(mappedFormData);
		}

		@Test
		void shouldCallRemoteService() {
			service.getFormData(messageMetaData);

			verify(remoteService).getMessage(XtaMessageTestFactory.MESSAGE_ID);
		}

		@Test
		void shouldCallMapper() {
			service.getFormData(messageMetaData);

			verify(mapper).toFormData(classification, messageMetaData, vorgangNummerSupplier);
		}

		@Test
		void shouldCallMapperToFormData() {
			service.getFormData(messageMetaData);

			verify(mapper).toFormData(classificationCaptor.capture(), messageMetaDataCaptor.capture(), eq(vorgangNummerSupplier));
			assertThat(messageMetaDataCaptor.getValue()).isEqualTo(messageMetaData);
			assertThat(classificationCaptor.getValue()).isEqualTo(classification);
		}

		@Test
		void shouldCallAddAttachments() {
			service.getFormData(messageMetaData);

			verify(service).addAttachments(message, mappedFormData);
		}

		@Test
		void shouldReturnMappedResult() {
			doReturn(mappedFormData).when(service).addAttachments(any(), any());

			var result = service.getFormData(messageMetaData);

			assertThat(result).isEqualTo(mappedFormData);
		}
	}

	@Nested
	class TestAddAttachments {

		private final FormData inFormData = FormDataTestFactory.createBuilder().clearAttachments().numberOfAttachments(0).build();

		@Test
		void shouldAddAttachments() {
			var result = service.addAttachments(XtaMessageTestFactory.create(), inFormData);

			assertThat(result.getAttachments()).hasSize(1);
			assertThat(result.getNumberOfAttachments()).isEqualTo(1);
		}

		@Test
		void shouldWorkWithoutAnyAttachment() {
			var result = service.addAttachments(XtaMessageTestFactory.createBuilder().clearAttachments().build(), inFormData);

			assertThat(result.getAttachments()).isEmpty();
			assertThat(result.getNumberOfAttachments()).isZero();
		}
	}

	@DisplayName("get representations attachments pair")
	@Nested
	class TestGetRepresentationsAttachmentsPair {

		@Mock
		private XtaMessageMetaData messageMetaData;

		@Mock
		private RepresentationsAttachmentsPair classification;

		private List<IncomingFile> incomingFiles;

		@BeforeEach
		void mock() {
			incomingFiles = List.of(IncomingFileTestFactory.create(), IncomingFileTestFactory.create());
		}

		@DisplayName("should use correct mapper xdomea message type")
		@Test
		void shouldUseCorrectMapperXdomeaMessageType() {
			when(messageMetaData.getMessageType()).thenReturn(XtaService.XDOMEA_0201_MESSAGE_TYPE);
			when(xdomeaMessageDataMapper.mapIncomingFilesToRepresentationsAttachmentsPair(incomingFiles))
					.thenReturn(classification);

			var result = service.getRepresentationsAttachmentsPair(messageMetaData, incomingFiles);

			assertThat(result).isEqualTo(classification);
		}

		@DisplayName("should use correct mapping for FIM message")
		@Test
		void shouldUseCorrectMappingForFimMessage() {
			when(messageMetaData.getMessageType()).thenReturn(FIM_MESSAGE_TYPE_PREFIX + "836487");
			doReturn(classification).when(service).mapIncomingFilesToRepresentations(incomingFiles);

			var result = service.getRepresentationsAttachmentsPair(messageMetaData, incomingFiles);

			assertThat(result).isEqualTo(classification);
		}

		@DisplayName("should throw exception for unexpected message type")
		@Test
		void shouldThrowExceptionForUnexpectedMessageType() {
			when(messageMetaData.getMessageType()).thenReturn("unexpected");

			assertThatThrownBy(() -> service.getRepresentationsAttachmentsPair(messageMetaData, incomingFiles))
					.isInstanceOf(TechnicalException.class);
		}
	}

	@DisplayName("map incoming files to representations")
	@Nested
	class TestMapIncomingFilesToRepresentations {

		private List<IncomingFile> incomingFiles;

		@BeforeEach
		void mock() {
			incomingFiles = List.of(IncomingFileTestFactory.create(), IncomingFileTestFactory.create());

		}

		@DisplayName("should return representations with incoming files")
		@Test
		void shouldReturnRepresentationsWithIncomingFiles() {
			var result = service.mapIncomingFilesToRepresentations(incomingFiles);

			assertThat(result.representations()).isEqualTo(incomingFiles);
		}

		@DisplayName("should return attachments with empty list")
		@Test
		void shouldReturnAttachmentsWithEmptyList() {
			var result = service.mapIncomingFilesToRepresentations(incomingFiles);

			assertThat(result.attachments()).isEmpty();
		}
	}

	@Nested
	class TestAcknowledgeReceive {

		@Test
		void shouldCallRemoteService() {
			service.acknowledgeReceive(XtaMessageTestFactory.MESSAGE_ID);

			verify(remoteService).close(XtaMessageTestFactory.MESSAGE_ID);
		}
	}

	@DisplayName("create XTA message stream")
	@Nested
	class TestCreateXtaMessageStream {
		private static final List<String> XTA_IDENTIFIERS = List.of("XTA_IDENTIFIER", "XTA_IDENTIFIER2");

		@Captor
		private ArgumentCaptor<String> identifierCaptor;

		@BeforeEach
		void mock() {
			when(remoteService.getMessagesMetadata())
					.thenReturn(XtaMessageMetaDatasAndHeaderTestFactory.create())
					.thenReturn(XtaMessageMetaDatasAndHeaderTestFactory.create());
			when(currentIdentifierService.getIdentifiers()).thenReturn(XTA_IDENTIFIERS);
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var result = service.createXtaMessageStream().toList();

			assertThat(result).isEqualTo(getExpectedMessages());
		}

		private List<XtaMessageMetaData> getExpectedMessages() {
			return Stream.concat(
					XtaMessageMetaDatasAndHeaderTestFactory.create().getMessages(),
					XtaMessageMetaDatasAndHeaderTestFactory.create().getMessages()).toList();
		}

		@DisplayName("should call set current identifier")
		@Test
		void shouldCallSetCurrentIdentifier() {
			service.createXtaMessageStream().toList();

			verify(currentIdentifierService, times(2))
					.setCurrentIdentifier(identifierCaptor.capture());
			assertThat(identifierCaptor.getAllValues()).isEqualTo(XTA_IDENTIFIERS);
		}
	}

}
